
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/semaphore.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/ext2_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/iobuf.h>
#include <linux/slab.h>
#include <linux/baio.h>
#include <linux/iobuf.h>

#define MAX_BAIO_PENDING 1024

extern struct buffer_head *reuse_list;
extern struct wait_queue *buffer_wait;

void free_async_buffers_baio(struct buffer_head *);

struct baio_entry * get_baio_entry()
{
	return kmalloc(sizeof(struct baio_entry), GFP_KERNEL| __GFP_BIGMEM);
}

struct baio_desc * get_baio_desc()
{
	return kmalloc(sizeof(struct baio_desc), GFP_KERNEL| __GFP_BIGMEM);
}

struct baio_queue * get_baio_queue_entry()
{
	return kmalloc(sizeof(struct baio_queue), GFP_KERNEL| __GFP_BIGMEM);
}

void put_baio_entry(struct baio_entry *ent)
{
	kfree(ent);
}

void put_baio_desc(struct baio_desc * ent)
{
	int i;
	if (!ent)
		return;
	for(i=0; i<ent->num_entries; i++) {
		if (ent->entry[i])
			kfree(ent->entry[i]);
	}
	if (ent->entry)
		kfree(ent->entry);
	if (ent)
		kfree(ent);
}

void put_baio_queue_entry(struct baio_queue *ent)
{
	if (ent)
		kfree(ent);
}

struct buffer_list* add_buffer_list(struct buffer_list *blist, struct buffer_list *node)
{
	struct buffer_list *tmp;

	//printk("\na2bl bl: %x, en: %x",blist,node);
	
	if (!blist) {
		blist = node;
		blist->next = blist->prev = NULL;
		//printk("\na2bl returns bl: %x", blist);
		return blist;
	}
	for(tmp = blist; tmp->next; tmp = tmp->next);
	tmp->next = node;
	tmp->next->next = NULL;
	tmp->next->prev = tmp;
	//printk("\na2bl returns bl: %x", blist);
	return blist;
}

void free_request_queue(struct baio_queue *request_queue)
{
	struct baio_queue *q = request_queue, *tmp;
	while(q) {
		tmp = q->next;
		kfree(q);
		q = tmp;
	}
}
		
		
struct baio_queue* add_to_request_queue(struct baio_queue *rq, struct baio_queue *qent)
{
	if (!qent)
		return rq;
	qent->next = rq;
	qent->prev = NULL;
	if (rq)
		rq->prev = qent;
	return qent;
	
}

int do_baio_rw(struct baio_desc *kdes)
{
	struct buffer_head *bh; int i;
	struct baio_entry *tmp;
	struct baio_struct *blk;
	struct buffer_list *buflist = NULL, *tmp_bl;
	struct baio_queue *q_entry;
	struct task_struct *tsk;
	status_t st;
	int to_wakeup = 0;
	
	if (!current->baio_slave) {
		printk("\nBAIO service daemon inoperative - aborting ..");
		return -EAGAIN;
	}

	tsk = current->baio_slave;

	blk = lookup_block_struct(kdes->blockfd);
	if (!blk)
	{
		printk("\nInvalid block descriptor");
		return -EINVAL;
	}
	
	//map logical to physical block nos. from info in per-process block
	//descriptor table (BDT)
	map_physical_blocks(blk, kdes);


	for(i = 0; i< kdes->num_entries; i++) {
		tmp = kdes->entry[i];
		tmp_bl = kmalloc(sizeof(struct buffer_list), GFP_KERNEL| __GFP_BIGMEM);
		if (!tmp_bl) {
			printk("\nerror in kmalloc"); return -EAGAIN;
		}
			
		st = IN_PROGRESS;
		put_user(st, tmp->status);

		if (kdes->flags & NOBUFFER) {
			int st = rw_raw_baio(tmp->op, blk, 
				tmp->data, blk->inode->sb->blocksize,
			       	tmp->block_nr, tmp_bl);
			if (st < 0) {
				printk("\nError in raw IO allocation");
				copy_to_user(tmp->status, &st, sizeof(status_t)); 
				return -ENOMEM;
			}
			bh = tmp_bl->buf;
		}
		else {
			bh = getblk (blk->inode->dev, tmp->block_nr, blk->inode->sb->blocksize);

			if (!bh) {
				printk("\ngetblk failed");
				st = ERROR;
				copy_to_user(tmp->status, &st, sizeof(status_t)); 
				return -ENOMEM;
			}

			if (tmp->op == BAIO_WRITE) {
				copy_from_user(bh->b_data, tmp->data, blk->inode->sb->blocksize);

				mark_buffer_dirty(bh, 1);
			}
			tmp_bl -> buf = bh;
			tmp_bl -> iobuf = NULL;
		}
		tmp_bl -> entry_num = i;
		tmp_bl -> op = tmp ->op;
		buflist = add_buffer_list(buflist, tmp_bl);
		if (tmp->op == BAIO_WRITE)
		{
		//printk("\nWrite: data is %c", bh->b_data[0]);
			ll_rw_block (WRITE, 1, &bh);
		}
		else
			ll_rw_block(READ, 1, &bh);
		tmp_bl->wait.task = current->baio_slave;
		bh->b_count ++;
		add_wait_queue(&bh->b_wait, &tmp_bl->wait);
	}
	run_task_queue(&tq_disk);

	q_entry = get_baio_queue_entry();
	if (!q_entry) {
		printk("\nError in kmalloc");
		return -EAGAIN;
	}
		
	q_entry->bl = buflist;
	q_entry->desc = kdes;
	
	//request queue goes from empty to non-empty; wakeup service daemon
	if (!tsk->request_queue)
	       to_wakeup = 1;	

	//atomically update the request queue 
	down(tsk->rq_sem);	
	//if (current->request_queue_lock == SPIN_UNLOCKED)
	//	printk("\nSurprise!! Locked and Unlocked");
	tsk->request_queue = add_to_request_queue(tsk->request_queue, q_entry);
	current->num_baio_pending ++;
	up(tsk->rq_sem);
	if (to_wakeup){
		//printk("baio_rw : waking up baio_service");	
		wake_up_process(tsk);
	}
	return 1;
}

asmlinkage int sys_baio( baio_desc *desc )
{
	struct baio_desc *kdes;
	struct baio_entry *tmp, **ent_arr;
	int i;

	if (current->num_baio_pending >= MAX_BAIO_PENDING) {
		printk("\nToo many baio requests");
		return -1;
	}
		
	kdes = get_baio_desc();
	if (!kdes){
		printk("\nOut of kernel memory");
		return -1;
	}
	
	copy_from_user(kdes, desc, sizeof(struct baio_desc));

	ent_arr = kmalloc((kdes->num_entries)* sizeof(baio_entry*), GFP_KERNEL|__GFP_BIGMEM);
	
	if (!ent_arr) {
		printk("\nOut of kernel memory");
		return -1;
	}

	//if ((access_ok(VERIFY_WRITE, kdes->entry, kdes->num_entries*sizeof(baio_entry *))))
		copy_from_user(ent_arr, kdes->entry, kdes->num_entries*sizeof(baio_entry *));
	/*else
		return -EFAULT;*/
	
	for(i=0; i< kdes->num_entries; i++) {
		tmp = get_baio_entry();
		if (!tmp) 
			return -1;
		copy_from_user(tmp, ent_arr[i], sizeof(struct baio_entry));
		ent_arr[i] = tmp;
		//printk("\nEntry %d: block_nr = %lu, user data addr = %x\n", i,ent_arr[i]->block_nr, ent_arr[i]->data);
				
	}
	kdes->entry = ent_arr;
	//printk("\nCalling do_baio_rw");
	return do_baio_rw(kdes);
}


int sys_set_baio_slave(int pid)
{
	struct task_struct *tsk = find_task_by_pid(pid);
	if (!tsk) {
		printk("\nInvalid pid !!!");
		return -EINVAL;
	}
	if (current->baio_slave) {
		printk("\nBaio slave already set");
		return -1;
	}
	current->baio_slave = tsk;
	tsk->rq_sem = kmalloc(sizeof(struct semaphore),GFP_KERNEL|__GFP_BIGMEM);
	if (!tsk->rq_sem) {
		printk("\nOut of kernel memory");
		return -1;
	}
	sema_init(tsk->rq_sem, 1);
	return 1;
}


void map_physical_blocks(struct baio_struct *blk, struct baio_desc *desc)
{
	int i;
	for(i = 0; i< desc->num_entries; i++) 
		desc->entry[i]->block_nr += blk->inode->start;
}


void update_user_data(struct baio_entry *ent, struct buffer_head *bh)
{
	status_t st = COMPLETE;
	//printk("\nUpdating user status");
	
	put_user(st, ent->status);

	//printk("\nI/O complete for block %lu", ent->block_nr);

	if (ent->op == BAIO_READ) {
		//printk("\nCopying data to user buffer address %x\n", ent->data);
		copy_to_user((void*)ent->data, (void*)bh->b_data, bh->b_size);
	}
}

struct buffer_list* remove_buffer_list(struct buffer_list *bl, struct buffer_list *d)
{
	struct buffer_list *tmp = bl;

	//printk("\nrm bl bl: %x, en: %x",bl,d);
	
	if (!bl)
		return NULL;
	
	for (; tmp ; tmp = tmp->next) {
		if (tmp == d) {
			if (tmp->prev) {
				tmp->prev->next = d->next;
				if (tmp->prev->next)
					tmp->prev->next->prev = tmp->prev ;
			}
			else {
				if (tmp != bl) {
					//printk("\nBuffer list corrupted"); 
					return NULL;
				}
					
				bl = tmp->next;
				if (bl)
					bl->prev = NULL;
				      
			}
			if (tmp->iobuf) {
				//cleanup_bounce_buffers(tmp->op, 1, &tmp->iobuf, 4096);
				unmap_kiobuf(tmp->iobuf);
				free_kiovec(1, &tmp->iobuf);
			}
			kfree(tmp);

			//printk("\nrm_bl returns bl: %x", bl);	
			return bl;
		}
	}
	printk("\nStrange !! Remove buffer called with non-existent buf");
	return bl;
}

void complete_baio_request(struct baio_queue *qent)
{
	status_t st;
	st = COMPLETE;
	copy_to_user(qent->desc->status, &st, sizeof(status_t));
	//send signal to the process qent->tsk 
	
	put_baio_desc(qent->desc);
}

struct baio_queue* remove_queue_entry(struct baio_queue *rq,struct baio_queue *qent)
{
	struct baio_queue *request_queue = rq;
	struct baio_queue *tmp = request_queue;

	//printk("\nrm_q rq: %x, en: %x",rq,qent);
	
	if (!rq)
		return NULL;
	
	for(; tmp; tmp = tmp->next) {
		if (tmp == qent) {
			if (tmp->prev) {
				tmp->prev->next = tmp->next;
				if (tmp->prev->next)
					tmp->prev->next->prev = tmp->prev;
			}
			else {
				if (tmp != request_queue){
					//printk("\nReq. queue corrupted");
					return NULL;
				}
				request_queue = tmp->next;
				if (request_queue)
					request_queue->prev = NULL;
			}
					
			put_baio_queue_entry(tmp);

			//printk("\nrm_rq returns rq: %x", request_queue);
			return request_queue;
		}
	}
	printk("\nSurprise !! Remove queue called with invalid queue entry\n");
	return request_queue;
}
	

	/* remove_buffer_list returns 0 if the list gets empty
	 * which means the baio request is complete and the process 
	 * can be notified through a signal if required
	 */
asmlinkage int sys_baio_service()
{
	struct baio_desc *desc;
	struct buffer_list *bl, *next_bl;
	struct baio_queue *q, *next_q;
	int sleep;
	struct task_struct *p = current->p_pptr;

	if (p->baio_slave != current) {
		printk("\nBaio service thread improperly set ..");
		return -1;
	}
	//p->baio_slave = current;
	down(current->rq_sem);

	while(1) {
repeat:
		if (!(current->request_queue)) {
			current->state = TASK_UNINTERRUPTIBLE;
			up(current->rq_sem);
			//printk("\nhmm.. queue empty. I'm going to sleep\n");
			schedule();
			//printk("\nSomeone woke me up!!!\n");

			if (current->killed) {
				//printk("\nOops ! someone killed me ..");
				return 1;
			}

			down(current->rq_sem);
			goto repeat ;
		}

		sleep = 0;
		//printk("\n\tGoing thro' req queue");
		
		q = next_q =  current->request_queue;

		if (current->killed) {
			//printk("\nOops! Some ass killed before rq");
			return 1;
		}

		while (next_q) {
			//printk("\n\t\t\tLooping inside rq");

			if (current->killed) {
				//printk("\nOops! Some ass killed inside rq");
				return 1;
			}
			
			q = next_q;
			next_q = q->next;
			desc = q->desc;
			bl = next_bl = q->bl;
			
			run_task_queue(&tq_disk);
			
			while(next_bl) {
				//printk("\n\t\t\t\t\tLooping inside bl");
				if (current->killed) {
					//printk("\nOops! Some ass killed inside bl");
					return 1;
				}

				bl = next_bl;
				next_bl = bl->next;
				
				if (!buffer_locked(bl->buf)) {
				        update_user_data(desc->entry[bl->entry_num], bl->buf);
					remove_wait_queue(&bl->buf->b_wait, &bl->wait);
					if (q->desc->flags & NOBUFFER)
						free_async_buffers_baio(bl->buf);
					else
						brelse(bl->buf);
					if ((q->bl = remove_buffer_list(q->bl, bl))==NULL) {
						complete_baio_request(q);
						current->request_queue = remove_queue_entry(current->request_queue, q);
						p->num_baio_pending --;
					}
				}
				else
					sleep = 1;
			}
		}

		if (sleep) {
			current->state = TASK_UNINTERRUPTIBLE ;
			up(current->rq_sem);
			//printk("\nGoing back to sleep");
			schedule();
			//printk("\nWaking up ..");
			if (current->killed) {
				//printk("\nOops ! someone killed me 2 ..");
				return 1;
			}
			down(current->rq_sem);
			//printk("\nend_buf_count is %lu", end_buf_count);
		}
	}
}


