#include <asm/siginfo.h> 

/* 	  BLOCK ASYNCHRONOUS I/O  
 * 	  List of syscalls that are provided to the user   */


#define MAX_NAMESIZE 64
#define MAX_BAIO_OPEN 32
#define BAIO_MAGIC (unsigned long)0x82a45b32c11d67ef
#define TRUE 1
#define FALSE 0

#define NOBUFFER 0x1

typedef enum status { IN_PROGRESS, COMPLETE, ERROR } status_t;
typedef enum op { BAIO_READ, BAIO_WRITE } op_t;
typedef void (*baio_notify_fn)(int);

typedef struct baio_entry {
	char *data;           //pointer to  user data buffer
	int shmid;	      //shared memory id	
	int block_nr;	      //logical block number in disk segment
	op_t op;
	status_t *status;      //per-block completion status  		 
}baio_entry;	
	
typedef struct baio_desc {
	int blockfd;	      // block descriptor
	baio_entry **entry;   // array of ptrs to baio_entries
	int num_entries;      // number of entries
	int flags;  	      // indicating mode of notification, etc.
	status_t *status;     // overall completion status	
	baio_notify_fn upcall;// upcall to the application
} baio_desc;


/* List of buffers used for a single baio request */
struct buffer_list {
	struct buffer_head *buf;
	int entry_num;			//to copy into appropriate user buffer
					//during reads, not used for write
	struct wait_queue wait;
	op_t op;
	struct buffer_list *prev, *next;
	struct kiobuf *iobuf;
};

struct baio_queue {
	struct baio_desc * desc;
	struct buffer_list *bl;
	struct baio_queue *prev, *next;
};





typedef unsigned long block_t;


struct baio_superblock {
	kdev_t dev;
	unsigned long magic;
	unsigned int blocksize; 
	block_t freemap_block;  //ptr to start block containing freelist
        block_t freeblock_hint;	
	block_t freemap_inode;  //ptr to start of inode freelist blocks
	block_t start_data;
	block_t start_inode;
	long nr_inodes;
	long nr_blocks;		//size of the device in blocks
};

struct baio_device_list {
	kdev_t dev;
	struct baio_superblock *sb;
	struct baio_device_list *next;
};


//The actual inode to be written to disk should exclude the last parameter
//i.e the size of an inode in disk is sizeof(struct baio_inode) - sizeof(sb);
//
struct baio_inode {
	unsigned int i_no;
	char dname[MAX_NAMESIZE];	 	//only short names (?)
	uid_t uid;
	gid_t gid;
	block_t start;		//starting block of disk segment
	unsigned int num_blocks;
	kdev_t dev;
	struct baio_superblock *sb;
};


struct baio_struct * lookup_block_struct(int);
void map_physical_blocks(struct baio_struct *, struct baio_desc *);

unsigned int allocate_blocks(struct baio_superblock *sb, int *nr_blocks, block_t *start_blocknr) ;
struct baio_inode * lookup_name(char *ds_name, struct baio_superblock *sb);
int install_inode(struct baio_inode *i_n);

/*	To create a disk segment in a specified device
 *	RETURNS
 *		1  on SUCCESS
 *		-1 on FAILURE, actual_nr contains available size
 *	devname  	-  pathname of the device file
 *	dname    	-  name of this disk segment
 *	nr_blocks	-  size of the disk segment required
 *	actual_nr	-  maximum size available
 */
 
int sys_create_dseg(char *devname, char *dname, int nr_blocks,int *actual_nr);

/* to delete a disk segment */
int sys_unlink_dseg(char *dev, char *dname);

/*	Opens a disk segment for access
 *	RETURNS
 *		A block descriptor to the disk segment, on SUCCESS
 *		-1 on FAILURE
 *	dname 		- name of the disk segment
 *	mode		- open mode 
 */	
int sys_open_dseg(char *dev, char *dname, int mode);
int sys_close_dseg(int blockfd);

/*	Writes/ Reads a set of blocks to/from an open disk segment
 *	 - desc contains complete info about the write/read
 */
int sys_baio(baio_desc *desc);

/*	This is the entry-point for the daemon that keeps track
 *	of pending baio requests and notifies applications on 
 *	completion.
 *	The daemon executes this syscall which never returns
 */
int sys_baio_service();
	
	
