/* 	
	Assumed Macro functions!

	HASH_BLK_NUM gives the logical block number in the series of blocks
	that holds the inodes - logical numbers start with a zero
	
	HASH_OFFSET gives the logical inode number within the series of inodes
	in the specific block - logical numbers start with a zero
	
	MOD(a,b)
*/

#define INODE_SIZE	(sizeof(struct baio_inode) -  sizeof(struct baio_superblock *))



#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/iobuf.h>
#include <asm/atomic.h>
#include <linux/slab.h>
#include <linux/baio.h>



typedef unsigned char bool;

int logb(int a)
{
	int i = 0;
	for(; a ; a = a >> 1, i++);
	if (i) 
		return i-1;
	else
		return i;
}

int MOD(int a, int b)
{
	if (b == 0)
		return 0;
	return a % b;
}


int CEIL(int nr, int dr) {


	
	if (!dr) {
		//printk("\nDivide by zero in CEIL");
		return nr;
	}
	if (nr == ((nr/dr)*dr)) 
		return (nr/dr);
       	else 
		return (nr/dr + 1);
}


int HASH(char *dsname, int max) {
	int	sum = 0;

	for(; *dsname ; sum += *dsname , dsname ++);
	return MOD (sum, max);
}




/*
	Function:	find_availability
	Author:		Venkatesh V
	Input:		free_byte - that has 1 byte freelist 
				free_bit - that on input shows the position where the
							inode wud be preferred
							on ouput has the actual position in which 
							the inode is allocated for this request
				On success, change the free_byte accordingly and return!
								
	Output:		On success ie a free inode available in the free_byte sent return 1
				else 0, if free_byte = 255!
*/

int		find_availability(char *free_byte, int *free_bit) {
		
		char	loop_test = 1;
		int	i=0;

		if (*free_bit > 7) return 0;

		*free_bit = 7 - *free_bit;
	
		loop_test = loop_test << *free_bit;

		if (!(*free_byte & loop_test)) {
			*free_byte += loop_test;
			*free_bit = logb(loop_test);
			*free_bit = 7 - (*free_bit);
			//printk("\nexiting from FA - success: %x, %x",*free_byte, *free_bit);
			return 1;
		}

		loop_test = ~(*free_byte);

		for(; loop_test; loop_test = loop_test >> 1, i++);

		if ( i == 0 ) {
			//printk("\nexiting from FA - failure");
			return 0;
		}

		loop_test = 1 << (i-1);
		
		*free_byte += loop_test;
		*free_bit = 8 - i;
		//printk("\nexiting from FA - miss & success");

		return 1;
	}




int		find_availability_old(char *free_byte, int *free_bit){

		char	test,
			loop_test=1;
		
		if (*free_bit > 7) return 0;
		
		*free_bit = 7 - *free_bit;
		
		test = loop_test = loop_test << *free_bit;
		
		do {
			//printk("\nFA:%x, %x",*free_byte, loop_test);
			if (!(*free_byte & loop_test)) {
				*free_byte += loop_test;
				*free_bit = logb(loop_test);
				*free_bit = 7 - (*free_bit);
				//printk("\nexiting from FA - success: %x, %x",*free_byte, *free_bit);
				return 1;
			}
		 		
			loop_test = (loop_test << 1);
			
			if (!loop_test) loop_test = 1;
			
		} while (loop_test != test);
		//printk("\nexiting FA failure");
		return 0;		
}
					

/* 	
	Function: 	allocate_bm
	Author:		Venkatesh V
	Input:		blk_no_req - Number of contiguos physical blocks requested
				start - Starting phisical block number, if allocated
						Zero, otherwise
	Output:		Actual number of available contigous physical blocks
	
	Assumptions: 	freeblock_hint as the nth logical data block number which is 
					likely to be free.....
*/

unsigned int	allocate_blocks(struct baio_superblock *sb, int *nr_blocks, block_t *start_blocknr) {

	block_t		start_blk_no,
			loop_blk_no,
			tgt_blk_no;
	
	int		start_offset,
			loop_offset,
			tgt_offset,
			max_avail,
			avail;
			
	bool		flag,
			exit_flag;
	
	
	struct	buffer_head *bh;

	start_blk_no = sb->freeblock_hint / (sb->blocksize * 8);
	
	start_offset = sb->freeblock_hint - (start_blk_no *  (sb->blocksize * 8));
	
	/*
	Paranoia
	*/
	
	if ( start_blk_no > ( sb->nr_blocks / (sb->blocksize * 8) ) ) {
		start_blk_no = 0;
	}
	
	if ( start_offset >= (sb->blocksize * 8) ) {
		start_offset = 0;
	}
	
	tgt_blk_no = loop_blk_no = start_blk_no;
	tgt_offset = loop_offset = start_offset = start_offset/8;	/* Only Byte level operations */

	max_avail = 0;
	avail = 0;
	exit_flag = 0;
	
	bh = getblk(sb->dev, sb->freemap_block + loop_blk_no, sb->blocksize);

	/* TO BE REMOVED */
	//printk("\nStarting freeblock search from block %lu", sb->freemap_block+loop_blk_no);
	/* END OF CODE TO BE REMOVED */
			
	if (!bh) {
		printk("\ngetblk failed in allocate_blocks");
		return 0;
	}

	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);


	do {
			
			flag = 0;
				
			while (!(*(bh->b_data + loop_offset))) {
			
				avail += 8;
					
				if (avail >= max_avail) max_avail = avail;
										
				if (max_avail >= *nr_blocks) {
					flag = 1;
					break;
				}
							
				if (++loop_offset >= bh->b_size) {
					flag = 2;
					break;
				}
				
				if ((exit_flag) && (loop_offset >= start_offset)) {
					flag = 4;
					loop_blk_no = -1;
					break;
				}
				
			}

			if (flag == 0)  {
				/* freelist broken */
				avail = 0;
				while (*(bh->b_data + loop_offset)) {
					loop_offset++;
					tgt_offset = loop_offset;
					if (loop_offset >= bh->b_size) {
						flag = 3;
						break;
					}
					
					if ((exit_flag) && (loop_offset >= start_offset)) {
						flag = 4;
						loop_blk_no = -1;
						break;
					}
				}
			}
			
			
			/*
			Paranoia
			
			if (exit_flag) {
				flag = 4;
				loop_blk_no = -1;
			}
			
			*/
			
			if (flag == 1) {
				/* success */
				break;
			}

			if ((flag == 2) || (flag == 3)) {
				/* got to resume in the next block if at all one exists */
				
				loop_blk_no = MOD((loop_blk_no + 1) , (sb->nr_blocks / (sb->blocksize * 8)));
				
				if ( (loop_blk_no==0) && (flag == 2)) {
					/* freelist is no more contiguous - re-initialize tgt ptrs*/
					avail = 0;	
				}
				
				/*
				Paranoia
				*/
				
				if (loop_offset < bh->b_size) {
					printk("\nloop_offset erroneous");
					return 0;
				}					
					
				loop_offset = 0;
				
				if (!avail) {
					/* flag == 3  || flag == 2 and looping around*/
					tgt_blk_no = loop_blk_no;
					tgt_offset = 0;
				}
				
				if (loop_blk_no == start_blk_no) {
				
					exit_flag = 1;
				}

				bh = getblk(sb->dev, sb->freemap_block + loop_blk_no, sb->blocksize);
	
				if (!bh) {
					printk("\ngetblk failed in allocate_blocks");
					return 0;
				}

				ll_rw_block(READ, 1, &bh);
				wait_on_buffer(bh);		

			}
			
	} while (loop_blk_no != -1);

	if (loop_blk_no == -1) {
	/* failure's exit point - godhelp  */
		*nr_blocks = max_avail;
		return 0;
	}

	if (flag == 1) { 
	/* success - toggle freelist and return starting logical blk number */
	
	/*
	Paranoia
	*/
	
		if ( max_avail < *nr_blocks ) {
			printk("\nVv's algo sucks in allocate_blocks");
			return 0;
		}
	
		if ( max_avail != avail) {
			printk("\nVv's algo sucks in allocate_blocks");
			return 0;
		}
		
		loop_blk_no = tgt_blk_no;
		loop_offset = tgt_offset;
	
		do {
	
			bh = getblk(sb->dev, sb->freemap_block + loop_blk_no, sb->blocksize);
			
			if (!bh) {
				printk("\ngetblk failed in allocate_blocks");
				return 0;
			}

			ll_rw_block(READ, 1, &bh);
			wait_on_buffer(bh);
			
			while (	avail > 0) {
				if (bh->b_data[loop_offset]) {
					printk("\nIncorrect manipulation");
					printk("\noptimistically Vv isn't doing the undo stuff");
					return 0;
				}
				else {
					bh->b_data[loop_offset] = 0xff;
					avail -= 8;
				}
				
				if (++loop_offset >= bh->b_size)
					break;
			}
			
			ll_rw_block(WRITE, 1, &bh);
			mark_buffer_dirty(bh,1);
			wait_on_buffer(bh);
			
			if (avail <= 0) {
				/* success's exit point - godhelp */
				//printk("\nExit via success");
				sb->freeblock_hint = ((tgt_blk_no * (sb->blocksize * 8)) + (tgt_offset * 8) + *nr_blocks);
				*start_blocknr =  ((tgt_blk_no * (sb->blocksize * 8)) + (tgt_offset * 8));
				return 1;
			}
			
			loop_blk_no++;

			if (loop_blk_no > (sb->nr_blocks / (sb->blocksize * 8))) 
				loop_blk_no = -1;
			
			loop_offset = 0;
			
		} while (loop_blk_no != -1);
	
	//printk("\nIn and out of success");
	//printk("\noptimistically Vv isn't doing the undo stuff");
	return 0;
	
	}

	return 0;
}

int is_equal(char *s1, char *s2)
{
	int i;
	for(i=0; i<MAX_NAMESIZE; i++) {
		if (s1[i] != s2[i])
			return 0;
	if (!s1[i])	
		return 1;
	}
	return 1;

}	


/*
	Function:	lookup_name
	Author:		Venkatesh V
	Input:		dsname - Name of the disk segment
				sb - Superblock of the file system
	Output:		Inode number of the requested name
	Assumptions:	Superblock size = Buffer blocksize	
*/

struct baio_inode * lookup_name(char *ds_name, struct baio_superblock *sb) {

	block_t		start_blk_no,
			blk_no	= HASH(ds_name, CEIL(sb->nr_inodes, sb->blocksize/INODE_SIZE));
	
	int		offset;
	
	int		loop_ptr,
			start_ptr;
	
	struct	baio_inode	*i_n,
				*tmp_inode;
	
	struct	buffer_head *bh;

	bool	last_blk_flag = 0;
	
	start_blk_no = blk_no;

	if (blk_no < (CEIL(sb->nr_inodes, sb->blocksize/INODE_SIZE) - 1))
		offset = HASH(ds_name, sb->blocksize/INODE_SIZE);
	else {
		int tmp = MOD(sb->nr_inodes, sb->blocksize/INODE_SIZE);
		last_blk_flag = 1;
		offset = HASH(ds_name, tmp);
	}
		
	if (blk_no > (sb->nr_inodes / (sb->blocksize / INODE_SIZE))) {
		printk("\nblk_no overflow in lookup_name");
		return NULL;
	}
	
	do {
	
		bh = getblk(sb->dev, sb->start_inode + blk_no, sb->blocksize);
	
		if (!bh ) {
			printk("\ngetblk failed in lookup_name");
			return NULL;
		}	
		
		ll_rw_block(READ, 1, &bh);
		wait_on_buffer(bh);
	
		start_ptr = loop_ptr = offset * INODE_SIZE;

		if (loop_ptr > sb->blocksize) {
			printk("\noffset overflow in lookup_name");
			return NULL;
		}
		
		do {
			if (loop_ptr + INODE_SIZE > bh->b_size)
				break;

			
			tmp_inode = (struct baio_inode *) (bh->b_data + loop_ptr);
			if (is_equal(tmp_inode->dname, ds_name)) {
		
				i_n = (struct baio_inode *)kmalloc(sizeof(struct baio_inode), GFP_KERNEL);
	
				if (!i_n) {
					printk("\nkmalloc failed in lookup_name");
					return NULL;
				}
	
				memcpy((void *)i_n, (void *)bh->b_data + loop_ptr, INODE_SIZE);
				return(i_n);
			
			}
			if (!last_blk_flag) 	
				loop_ptr = MOD((loop_ptr + INODE_SIZE), INODE_SIZE * (bh->b_size / INODE_SIZE));
			else
				loop_ptr = MOD((loop_ptr + INODE_SIZE), INODE_SIZE * MOD(sb->nr_inodes, sb->blocksize/INODE_SIZE));

		} while(loop_ptr != start_ptr);	
		
		blk_no = MOD((blk_no+1), CEIL(sb->nr_inodes, (bh->b_size / INODE_SIZE)));

	if (blk_no == (CEIL(sb->nr_inodes, (bh->b_size / INODE_SIZE)) - 1 ))
		last_blk_flag = 1;
	else
		last_blk_flag = 0;

		
	} while(blk_no != start_blk_no);
	
	//printk("\nExiting name_lookup on failure");
	return NULL;
	
}


/*
	Function:	install_inode
	Author:		Venkatesh V
	Input:		i_n - the inode to be installed
	Output:		Success :  1
				Failure : -1
	Assumptions:	Superblock size = Buffer blocksize
*/

int		install_inode(struct baio_inode *i_n){

	block_t		actual_blk_num,
			actual_seq_num,
			start_blk_num;
	
	int		start_ptr,
			start_bit,
			loop_ptr,
			loop_bit,
			hit;
	bool		last_blk_flag=0;
			
	char	buf_byte;
			
	struct 	baio_superblock *sup_blk = i_n->sb;
	
	int 	blk_no	= HASH(i_n->dname, CEIL(sup_blk->nr_inodes, (sup_blk->blocksize / INODE_SIZE)));
	int 	offset;
	
	struct 	buffer_head *bh;
	if (blk_no < (CEIL(sup_blk->nr_inodes, sup_blk->blocksize/INODE_SIZE) - 1))
		offset = HASH(i_n->dname, sup_blk->blocksize/INODE_SIZE);
	else {
		int tmp =MOD(sup_blk->nr_inodes, sup_blk->blocksize/INODE_SIZE);
		last_blk_flag = 1;
		offset = HASH(i_n->dname, tmp);
	}

	//printk("\nblk_no = %d, offset = %d", blk_no, offset);	
	actual_blk_num = blk_no * (sup_blk->blocksize / INODE_SIZE);
	
	actual_seq_num = actual_blk_num += offset;
	//printk("\nactual seq_num = %lu", actual_seq_num);
	
	start_blk_num = actual_blk_num = actual_blk_num / (sup_blk->blocksize * 8);
	
	actual_seq_num -= actual_blk_num * sup_blk->blocksize * 8;
	
	
	do {
	
		bh = getblk(sup_blk->dev, sup_blk->freemap_inode + actual_blk_num, sup_blk->blocksize);

		if (!bh) {
			printk("\ngetblk failed in install_inode");
			return -ENOMEM;
		}
	
		ll_rw_block(READ, 1, &bh);
		wait_on_buffer(bh);

		loop_ptr = start_ptr = actual_seq_num / 8;
	
		loop_bit = start_bit = actual_seq_num - (start_ptr * 8);
		//printk("\nloop_ptr = %d, loop_bit = %d", loop_ptr, loop_bit);	
   
	    do {
    
   		buf_byte = *(bh->b_data + loop_ptr);
    
    		hit = 0;
    	
	    	if (find_availability(&buf_byte, &loop_bit)){
	    		memcpy(bh->b_data + loop_ptr, &buf_byte, 1);
	    		
	    		ll_rw_block(WRITE, 1, &bh);
	    		mark_buffer_dirty(bh, 1);
	    		wait_on_buffer(bh);
	    			    		
    			hit = 1;	
    			break;
	    	}
    	
  	    	if (!last_blk_flag)
			loop_ptr = MOD((loop_ptr + 1), bh->b_size);
		else {
			int tmp = MOD(sup_blk->nr_inodes, sup_blk->blocksize * 8);
			
			loop_ptr = MOD((loop_ptr + 1), tmp/8); 
		}
    	
    		loop_bit = start_bit;			/*Cud be made zero too, but I hope this is more efficient! */
 	
	    } while ( loop_ptr != start_ptr);
    
       	if (hit) break;    	
    	
    	actual_blk_num = MOD((actual_blk_num + 1), CEIL(sup_blk->nr_inodes, sup_blk->blocksize * 8) );
    	
    	/*actual_seq_num = 0;				/ *Cud be made zero too, but I hope this is more efficient! */
    
	if (actual_blk_num == (CEIL(sup_blk->nr_inodes, sup_blk->blocksize * 8) - 1))	last_blk_flag = 1;
	else	
		last_blk_flag = 0;
	
    } while (actual_blk_num != start_blk_num);
    
    if (!hit) {
	//printk("\nExiting install_inode with error (!hit)");
	return -1;
    }
        
    actual_seq_num = (loop_ptr * 8) + loop_bit;
    
    actual_seq_num += actual_blk_num * sup_blk->blocksize * 8;
		
	blk_no = actual_seq_num / (sup_blk->blocksize / INODE_SIZE);
	
	offset = actual_seq_num - (blk_no * (sup_blk->blocksize / INODE_SIZE ));
	
	offset *= INODE_SIZE;
	
	bh = getblk(sup_blk->dev, sup_blk->start_inode + blk_no, sup_blk->blocksize);
		
	if (!bh ) {
		printk("\ngetblk failed in install_inode");
		return -ENOMEM;
	}
	
	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	//printk("\nWriting inode of dseg name : %s : to disk");
	
	memcpy(bh->b_data + offset, i_n, INODE_SIZE);
	
	ll_rw_block(WRITE, 1, &bh);
	mark_buffer_dirty(bh,1);
	wait_on_buffer(bh);
 	//printk("\nSuccess exit from install_inode");	
	return(1);
	
}
