#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/ext2_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/iobuf.h>
#include <asm/atomic.h>
#include <linux/slab.h>
//temporary
#include <linux/baio.h>

#define MAX_NAMESIZE 64
#define MAX_BAIO_OPEN 32
#define BAIO_MAGIC (unsigned long)0x82a45b32c11d67ef

//the global baio mount list
struct baio_device_list *baio_dev;

struct baio_superblock * get_baio_superblock()
{
	return kmalloc(sizeof(struct baio_superblock), GFP_KERNEL);
}

struct baio_inode * get_baio_inode()
{
	return kmalloc(sizeof(struct baio_inode), GFP_KERNEL);
}

void put_baio_superblock(struct baio_superblock * sb)
{
	kfree(sb);
}

void put_baio_inode(struct baio_inode *inode)
{
	kfree(inode);
}


void alloc_task_blocks(struct task_struct *p)
{
	int i;
	if (p->blocks) {
		printk("\nAlloc_task_block called for already allocated proc");
		return;
	}
	p->blocks = kmalloc(sizeof(struct baio_info), GFP_KERNEL);
	for(i =0; i< MAX_BAIO_OPEN; i++)
		p->blocks->bd_list[i] = NULL;
	p->blocks->next_bd = 0;
}

int install_bd(struct baio_struct *b)
{
	struct task_struct *p = current;
	int bd;
	if (!p->blocks)
		alloc_task_blocks(p);
	//TO BE REMOVED
	else {
		printk("\nStrange !!! process' blocks already allocated");
		return 1;
	}
	//TO BE REMOVED
		
	bd = p->blocks->next_bd;
	p->blocks->next_bd = bd+1;
	p->blocks->bd_list[bd] = b;
	return bd;
}

struct baio_struct* lookup_block_struct(int bd)
{
	if (current->blocks && bd <= MAX_BAIO_OPEN)
		return current->blocks->bd_list[bd];
	else
		return NULL;
}

struct baio_superblock * lookup_baio_superblock(kdev_t dev)
{
	struct baio_device_list *tmp = baio_dev;
	if (!tmp)
		return NULL;
	for(; tmp ; tmp = tmp->next) {
		if (tmp->dev == dev)
			return tmp->sb;
	}
	return NULL;
}

void add_baio_superblock(kdev_t dev, struct baio_superblock *sb)
{
	struct baio_device_list *tmp = baio_dev;
	if (!baio_dev) {
		baio_dev = kmalloc(sizeof(struct baio_device_list), GFP_KERNEL);
		baio_dev->dev = dev;
		baio_dev->sb = sb;
		baio_dev->next = NULL;
		return;
	}
	for(; tmp->next ; tmp = tmp->next) ;
	tmp->next = kmalloc(sizeof(struct baio_device_list), GFP_KERNEL);
	tmp->next->dev = dev;
	tmp->next->sb = sb;
	tmp->next->next = NULL;
}



int create_disk_segment(char *dev,char *dname, int nr_blocks, int *actual_nr)
{
	struct dentry *dentry;
	struct baio_superblock *sb;
	struct baio_inode *inode;
	block_t start;
	if (!dev)
		return -EFAULT;
	dentry = open_namei(dev, 0, 0);
	sb = lookup_baio_superblock(dentry->d_inode->i_rdev);
	if (!sb) {
		printk("\nThis device is not configured for baio");
		return -1;
	}
	*actual_nr = nr_blocks;
	printk("\nCreate: sb->nr_blocks =%lu, sb->dev=%d",sb->nr_blocks, sb->dev);

/*	if (lookup_name(dname, sb)) {
		printk("\nA disk segment with the name already exists");
		return -EINVAL;
	}*/

	if (! allocate_blocks(sb,  actual_nr, &start))
		return -EAGAIN;
	//printk("\nNew freehint val = %lu", sb->freeblock_hint);

	inode = get_baio_inode();
	memcpy(inode->dname, dname, MAX_NAMESIZE);        
	inode -> start = start + sb->start_data;
	inode -> num_blocks = nr_blocks;
	inode -> uid = current -> uid;
	inode -> gid = current -> gid;
	inode -> dev = sb -> dev;
	inode -> sb  = sb;
	//printk("\nTrying to install inode .. \n");	
	if (install_inode( inode ) < 0) {
		printk("\nError allocating new inode for the disk block");
		return -EAGAIN;
	}

	return 1;
}


asmlinkage int sys_create_dseg(char *dev, char *dname, int nr_bl, int *a_nr){
	/*char *c;
	sys_shmat(nr_bl, 0, 0, &c);
	printk("\nShared Memory: %c%c%c", c[0], c[1], c[2]);
	return 1;*/
	char *k_devname = getname(dev);
	char *k_dname = getname(dname);
	int actual;

	int retval = create_disk_segment(k_devname, k_dname, nr_bl, &actual);
	copy_to_user(a_nr, &actual, sizeof(int));
	putname(k_devname);
	putname(k_dname);
	return retval;
}
	


struct baio_superblock * baio_read_super(kdev_t dev)
{
	struct baio_superblock *sb;
	struct buffer_head *bh;
	
	sb = get_baio_superblock();
	printk("\nRead_super called for device %d", dev);

	bh = getblk(dev, 0, blksize_size[MAJOR(dev)][MINOR(dev)]);

	if (!bh) 
		return NULL;

	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);

	memcpy((char*)sb, bh->b_data, sizeof(struct baio_superblock));
	brelse(bh);
	return sb;
}

int baio_write_super(struct baio_superblock *sb)
{
	struct buffer_head *bh;
	
	bh = getblk(sb->dev, 0, blksize_size[MAJOR(sb->dev)][MINOR(sb->dev)]);
	if (!bh) 
		return -EAGAIN;

	ll_rw_block(WRITE, 1, &bh);
	mark_buffer_dirty(bh, 1);
	wait_on_buffer(bh);
	brelse(bh);
	return 1;
}


int open_disk_segment(char *dev, char *dname, int flags)
{
	struct baio_inode *inode;
	struct baio_struct *b;
	struct dentry *dentry;
	struct baio_superblock *sb;

	dentry = open_namei(dev, 0, 0);
	sb = lookup_baio_superblock(dentry->d_inode->i_rdev);
	if (!sb) {
		printk("\nBAIO system not configured for this device");
		return -EINVAL;
	}
	
	inode = lookup_name(dname, sb);
	if (!inode) {
		printk("\nNon-existent disk segment");
		return -EINVAL;
	}
	inode->sb = sb;

	b = kmalloc(sizeof(struct baio_struct), GFP_KERNEL);

	b->flags = flags;
	b->inode = inode;
	atomic_set(&b->count, 1);
	return install_bd(b);
}


asmlinkage int sys_open_dseg(char *dev, char *dname, int flags)
{
	char *k_dname = getname(dname);
	char *k_dev = getname(dev);
	int retval = open_disk_segment(k_dev, k_dname, flags);
	putname(dname);
	putname(dev);
	return retval;
}

int close_disk_segment(int bd)
{
	struct baio_struct *b ;
	if (current->blocks && bd < MAX_BAIO_OPEN) {
		b = current->blocks->bd_list[bd];
		current->blocks->bd_list[bd] = NULL;
		if (atomic_dec_and_test(&b->count))  // ! required ??? to be tested
			kfree(b);
		return 1;
	}
	return 0;
}


asmlinkage int sys_close_dseg(int bd)
{
	return close_disk_segment(bd);
}


int baio_mount (char *dname)
{
	struct baio_superblock *sb;
	struct dentry *dentry;
	kdev_t dev;
	if (!dname)
		return -EFAULT;
	dentry = open_namei(dname, 0, 0);
	dev = dentry->d_inode->i_rdev;

	if ((sb = lookup_baio_superblock(dev))) {
		printk("\nDevice already mounted for BAIO");
		return 1;
	}
	sb = baio_read_super(dev);


	if (sb->magic != BAIO_MAGIC) {
		printk("\nDisk is not configured for baio");
		return -EAGAIN;
	}
	//sb->blocksize = get_hardblocksize(dev);
	set_blocksize(dev, sb->blocksize);

	add_baio_superblock(dentry->d_inode->i_rdev, sb);
	return 1;
}


asmlinkage int sys_baio_mount(char *dname)
{
	char *k_dname = getname(dname);
	int retval = baio_mount(k_dname);
	putname(dname);
	return retval;
}


