#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <asm/unistd.h>
#include <errno.h>
#include <stdio.h>

typedef unsigned long block_t;
typedef unsigned short kdev_t;

#define BLOCKSIZE 4096 
#define MAX_NAMESIZE 64
#define BAIO_MAGIC (unsigned long) 0x82a45b32c11d67ef

_syscall2(int, my_write, int, blocknum, char*, buf)

struct baio_superblock {
	kdev_t dev;
	unsigned long magic;
	unsigned int blocksize; 
	block_t freemap_block;  //ptr to start block containing freelist
        block_t freeblock_hint;	
	block_t freemap_inode;  //ptr to start of inode freelist blocks
	block_t start_data;
	block_t start_inode;
	long nr_inodes;
	long nr_blocks;		//size of the device in blocks
};

struct baio_inode {
	unsigned int i_no;
	char dname[MAX_NAMESIZE];	 	//only short names (?)
	uid_t uid;
	gid_t gid;
	block_t start;		//starting block of disk chunk
	unsigned int num_blocks;
	kdev_t dev;
	struct baio_superblock *sb;
};



int ceil(float a)
{
	if (a > (int)a )
		return ((int)a + 1);
	else
		return (int)a;
}

/* Creates a baio system on the device specified
 * Organization of data in disk is as follows :
 * 	 Superblock structure (1 block)
 * 	 Inode Freelist bitmap blocks
 * 	 Inode blocks
 * 	 Data Freelist bitmap blocks
 * 	 Data blocks
 */ 	

void usage()
{
	fprintf(stderr, "Usage:  mkbaio dev-filename \n");
}

main(int argc, char **argv)
{
	int fd;
	long nr_blocks;
	long nr_inodes;
	long nr_inode_freelist_blocks;
	long nr_inode_blocks;
	long nr_data_freelist_blocks;
	long nr_inode_per_block;
	long nr_remain, nr_data_blocks;
		
	struct baio_superblock sb;
	struct stat st;
	char *all_zeroes = (char*)malloc(BLOCKSIZE);
	int i;

	memset(all_zeroes, 0, BLOCKSIZE);

	if (argc != 2)  {
		usage();
		return -1;
	}
	fd = open(argv[1], O_RDWR);
	if (fd < 0) {
		printf("\nBad device name "); return;
	}
 	ioctl(fd, BLKGETSIZE, &nr_blocks);


	stat(argv[1],  &st);
	if (st.st_rdev == 769) {
		printf("\nTrying to format primary partition !!! Aborting !");
		return -1;
	}

	nr_blocks /= (BLOCKSIZE/512);  //get no. of blocks from nr_sect

	printf("\nTotal No of blocks  = %d\n", nr_blocks);
	//fixing 1MB as the least possible avg. size of a block, so #inodes = 
	//#blocks / 1000
	nr_inodes = nr_blocks/1000;

	nr_inode_freelist_blocks = ceil((float)nr_inodes/(8* BLOCKSIZE));
	nr_inode_per_block = BLOCKSIZE/sizeof(struct baio_inode);
	nr_inode_blocks = ceil((float)nr_inodes / nr_inode_per_block);

	nr_remain = nr_blocks - nr_inode_freelist_blocks - nr_inode_blocks -1;
	
	nr_data_freelist_blocks = ceil((float)nr_remain / (8*BLOCKSIZE));
	
	nr_data_freelist_blocks -= nr_data_freelist_blocks/(8*BLOCKSIZE) ;
	
	nr_data_blocks = nr_remain - nr_data_freelist_blocks;

	sb.start_inode = nr_inode_freelist_blocks + 1;
	sb.start_data = sb.start_inode + nr_inode_blocks + nr_data_freelist_blocks;
	sb.freemap_inode = 1;
	sb.freemap_block  = sb.start_inode + nr_inode_blocks;
	sb.freeblock_hint = sb.freemap_block;
	sb.dev = st.st_rdev;
	sb.blocksize = BLOCKSIZE;	
       	sb.nr_blocks = nr_data_blocks;
	sb.nr_inodes = nr_inodes;
	sb.magic = BAIO_MAGIC;
	printf("\nNo. of data blocks = %d, No. of inodes = %d\n", nr_data_blocks, sb.nr_inodes);

	write(fd, (char*)&sb, sizeof(struct baio_superblock));

	/* initialize freemap blocks with all zeroes's */
	for(i= 1; i< sb.start_data + 1; i++)
		write(fd, all_zeroes, BLOCKSIZE);
	close(fd);
	sync();
		
}

	
	
	
       	


		
